import json
from http import HTTPStatus
import logging

from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

logger = logging.getLogger(__name__)


class BucketPathRewriterMiddleware(BaseHTTPMiddleware):
    """
    Rewrites a specified part of the request path with bucket name.

    Parameters
    ----------
    app : ASGIApp
        Application instance the middleware is being registered to.
    """

    def __init__(self, app: ASGIApp) -> None:
        """
        Initializes the BucketPathRewriterMiddleware.
        """
        super().__init__(app, self.dispatch)

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        """
        Rewrites a specified part of the request path with bucket name.

        Parameters
        ----------
        request : Request
            The incoming request.
        call_next : RequestResponseEndpoint
            Next middleware or endpoint on the execution stack.

        Returns
        -------
        Response
            The response generated by the middleware.
        """
        if request.method != "POST":
            return await call_next(request)

        body = await request.body()

        try:
            body = json.loads(body)
        except json.JSONDecodeError:
            content = {"message": "Invalid Request"}
            status_code = HTTPStatus.BAD_REQUEST

            return Response(content=json.dumps(content), status_code=status_code.value)

        logger.debug(f"Received body: {body}")

        bucket_name = body["bucket"]

        request.scope["path"] = "/" + bucket_name

        return await call_next(request)
